#ifndef OT_LAYOUT_GPOS_EXTENSIONPOS_HH
#define OT_LAYOUT_GPOS_EXTENSIONPOS_HH

namespace OT {
namespace Layout {
namespace GPOS_impl {

struct ExtensionPos : Extension<ExtensionPos>
{
  typedef struct PosLookupSubTable SubTable;
};

}
}
}

#endif /* OT_LAYOUT_GPOS_EXTENSIONPOS_HH */
